@@ -5,6 +5,17 @@
 #else
 #include <sys/types.h>
 #include <unistd.h>
+#include <stdio.h>
+#include <stdarg.h>
+#endif
+#ifdef __MACH__
+#include <sys/sysctl.h>
+#include <mach-o/fat.h>
+#include <mach-o/loader.h>
+#include <sys/stat.h>
+#include <sys/wait.h>
+#include <stdlib.h>
+#include <string.h> 
 #endif
 
 #include "mktmpdir.c"
@@ -75,6 +86,30 @@ int extract_embedded_file(embedded_file_t *emb_file, const char* ext_name, const
     return EXTRACT_OK;
 }
 
+/* Simple magic number reader for MacOS */
+#ifdef __MACH__
+uint32_t read_magic(FILE *obj_file, int offset) {
+  uint32_t magic;
+  fseek(obj_file, offset, SEEK_SET);
+  fread(&magic, sizeof(uint32_t), 1, obj_file);
+  return magic;
+}
+
+/* double slashes in the tmpdir path confuse execve on macOS */
+char *sanitise_tmp(char *s) {
+  if (s) {
+    const char *src = s;
+    char *dst = s;
+    while ((*dst = *src) != '\0') {
+      do {
+        src++;
+      } while (*dst == '/' && *src == '/');
+      dst++;
+    }
+  }  
+  return s;
+}
+#endif
 
 /* turn off automatic globbing of process arguments when using MingW */
 #if defined(WIN32) && defined(__MINGW32__)
@@ -158,6 +193,16 @@ char* shell_quote(const char *src)
 }
 #endif
 
+void die(const char *format, ...)
+{
+    va_list ap;
+    va_start(ap, format);
+    vfprintf(stderr, format, ap);
+    va_end(ap);
+    
+    exit(255);
+}
+
 char pp_version_info[] = "@(#) Packed by PAR::Packer " PAR_PACKER_VERSION;
 
 int main ( int argc, char **argv, char **env )
@@ -178,28 +223,93 @@ typedef BOOL (WINAPI *pALLOW)(DWORD);
 #endif
 #endif
 
-#define DIE exit(255)
-
     par_init_env();
 
-    stmpdir = par_mktmpdir( argv );	
-    if ( !stmpdir ) DIE;        /* error message has already been printed */
+    stmpdir = par_mktmpdir( argv );
+    if ( !stmpdir ) die("");        /* error message has already been printed */
 
     rc = my_mkdir(stmpdir, 0700);
     if ( rc == -1 && errno != EEXIST) {
-	fprintf(stderr, "%s: creation of private cache subdirectory %s failed (errno= %i)\n", 
-                        argv[0], stmpdir, errno);
- 	DIE;
+	die("%s: creation of private cache subdirectory %s failed (errno= %i)\n", 
+            argv[0], stmpdir, errno);
     }
 
     /* extract embedded_files[0] (i.e. the custom Perl interpreter) 
      * into stmpdir (but under the same basename as argv[0]) */
-    my_prog = par_findprog(argv[0], strdup(par_getenv("PATH")));
+    my_prog = par_findprog(argv[0], par_getenv("PATH"));
+
+#ifdef __MACH__
+    {
+        /* Detect if FAT binary */
+        FILE *obj_file = fopen(my_prog, "rb");
+        uint32_t magic = read_magic(obj_file, 0);
+        fclose(obj_file);
+
+        if (magic == FAT_CIGAM || magic == FAT_MAGIC) 
+        {
+            /* Create separate dir for extracted thin binary*/
+            char *ftmpdir = malloc(strlen(stmpdir) + 5 + 1);
+            sprintf(ftmpdir, "%s%s", stmpdir, "/thin");
+            sanitise_tmp(ftmpdir);
+            rc = my_mkdir(ftmpdir, 0700);
+            if (rc == -1 && errno != EEXIST)
+                die("%s: creation of cache subdirectory "
+                    "for extracted macOS thin binary %s failed (errno= %i)\n", 
+                    argv[0], ftmpdir, errno);
+            
+            /* Get architecture name */
+            size_t size;
+            sysctlbyname("hw.machine", NULL, &size, NULL, 0);
+            char *arch = malloc(size);
+            sysctlbyname("hw.machine", arch, &size, NULL, 0);
+
+            /* Detect if lipo exists, if not die */
+            struct stat buffer;
+            int exist = stat("/usr/bin/lipo", &buffer);
+            if (exist == -1)
+                die("%s: cannot find /usr/bin/lipo to unpack universal binary - "
+                    "do you need install Developer Tools? (errno=%i)\n", 
+                    argv[0], errno);
+          
+            char *archthinbin = malloc(strlen(ftmpdir) + 1 + strlen(par_basename(my_prog)) + 1);
+            sprintf(archthinbin, "%s/%s", ftmpdir, par_basename(my_prog));
+            char* lipo_argv[] = { "lipo", "-extract_family", arch, "-output", archthinbin, my_prog, NULL };
+            pid_t pid = fork();
+            if (pid == -1) 
+                die("%s: fork failed (errno=%i)\n",  argv[0], errno);
+            if (pid == 0)
+            {
+                /* child */
+                execve("/usr/bin/lipo", lipo_argv, env);
+                exit(1);
+            }
+
+            /* parent */
+            int wstatus;
+            waitpid(pid, &wstatus, 0);
+            if (!(WIFEXITED(wstatus) && WEXITSTATUS(wstatus) == 0))
+                die("%s: extracting %s binary with lipo failed (wstatus=%i)\n",
+                    argv[0], arch, wstatus);
+            free(arch);
+
+            /* exec correct thin binary */
+            exist = stat(archthinbin, &buffer);
+            if (exist == -1)
+                die("%s: cannot find thin binary %s to run (errno=%i)\n", 
+                    argv[0], archthinbin, errno);
+
+            argv[0] = archthinbin;
+            execve(archthinbin, argv, env);
+            die("%s: cannot execute thin binary %s (errno=%i)\n", 
+                argv[0], archthinbin, errno);
+        }
+    }
+#endif    
+    
     rc = extract_embedded_file(embedded_files, par_basename(my_prog), stmpdir, &my_perl);
     if (rc == EXTRACT_FAIL) {
-        fprintf(stderr, "%s: extraction of %s (custom Perl interpreter) failed (errno=%i)\n", 
-                            argv[0], my_perl, errno);
-        DIE;
+        die("%s: extraction of %s (custom Perl interpreter) failed (errno=%i)\n", 
+            argv[0], my_perl, errno);
     }
 
     if (rc == EXTRACT_OK)       /* i.e. file didn't already exist */
@@ -239,9 +349,8 @@ typedef BOOL (WINAPI *pALLOW)(DWORD);
     emb_file = embedded_files + 1;
     while (emb_file->name) {
         if (extract_embedded_file(emb_file, emb_file->name, stmpdir, &my_file) == EXTRACT_FAIL) {
-            fprintf(stderr, "%s: extraction of %s failed (errno=%i)\n", 
-                                argv[0], my_file, errno);
-            DIE;
+            die("%s: extraction of %s failed (errno=%i)\n", 
+                argv[0], my_file, errno);
         }
         emb_file++;
     }
@@ -278,7 +387,8 @@ typedef BOOL (WINAPI *pALLOW)(DWORD);
     exit(rc);
 #else
     execvp(my_perl, argv);
-    DIE;
+    die("%s: exec of %s (custom Perl interpreter) failed (errno=%i)\n", 
+        argv[0], my_perl, errno);
 #endif
 }
 
